' *******************************************************************************************************
' *                                                                                                     *
' *     AiChip_JVM_Demo_002.spin                                                                        *
' *                                                                                                     *
' *******************************************************************************************************

PRI Version
  return String("AiChip_JVM_Demo_002")

CON
{

As supplied the jem.out is ...

'        import stamp.*;
'        public class testJVM
'        {
'          static int iStatic;
'          static int incr( int a )
'          {
'            return a + 1;
'          }
'          static void main()
'          {
'            iStatic = 0;
'            while (true)
'            {
'              iStatic = java.lang.Math.min( incr(iStatic), 10);
'              stamp.core.CPU.delay(iStatic);
'            }
'          }
'        }

Progress so far ...

        DONE - Get 'main' running by itself using local variables
        DONE - Get 'main' calling other class native methods
        DONE - Get 'main' calling same class static method
        DONE - Get 'main' calling other class static methods
        DONE - Get 'main' called from program initialisation
        DONE - Create simple heap handler
        DONE - Handle array creation
        DONE - Handle class static variables for 'testJVM'
        DONE - Get main running with class static variables for 'testJVM'

        TODO - Add Native Method handling code
        
Not Implemented yet ...

        Jem_INVOKEVIRTUAL
        Jem_INVOKEINTERFACE
        Jem_ATHROW:       
        Jem_CHECKCAST
        Jem_INSTANCEOF
        Jem_WIDE
        Jem_GETFIELD
        Jem_PUTFIELD
        Jem_NEW
        Jem_ANEWARRAY
        Jem_LDC
        Jem_LDC_W
        Jem_LDC2_W
        Jem_TABLESWITCH
        Jem_LOOKUPSWITCH

        PopObjectRef            Need for IF_ACMPEQ/NE
      
Not completely implemented ...

        InvokeProgram           Just grabs an array for static fields
        Jem_NEWARRAY            What does the operand mean ? Always $00 !

}
CON
{

PLEASE NOTE : If modifying and re-releasing this code, please change the prefix of the source file from
              AiChip so as to avoid any confusion as to the origin of the source and update all internal
              references as appropriate. If you wish to credit the orginal source, anything  along the
              lines of, "Based on original AiChip_Jvm_XXX from AiChip Industries", is acceptable.

              AiChip is a trademark of AiChip Industries.

}
CON

' .-----------------------------------------------------------------------------------------------------.
' |     Board Configuration                                                                             |
' `-----------------------------------------------------------------------------------------------------'

  _CLKMODE            = XTAL1 + PLL16x                  ' 80MHz System Clock Speed
  _XINFREQ            = 5_000_000

OBJ

  'tv                  : "TV_Text"                       ' TV Text Display used for debugging
  tv                  : "Jazzed_JVM_MonDebug"           ' TV Text Display used for debugging
  jem                 : "AiChip_JVM_JemFile_002.spin"   ' JEM Bytecode to execute ( jem.bin )
  jzl                 : "Jazzed_JVM_jazzLib"            ' Library of jemcode methods from Jazzed
  pvl                 : "Jazzed_JVM_peterLib"           ' Library of jemcode and native methods from Peter

CON

' .-----------------------------------------------------------------------------------------------------.
' |     Hardware Configuration                                                                          |
' `-----------------------------------------------------------------------------------------------------'

  TV_PIN              = 12                              ' Base TV Pin

CON

' .-----------------------------------------------------------------------------------------------------.
' |     JVM Configuration                                                                               |
' `-----------------------------------------------------------------------------------------------------'

  INT32               = false                           ' Set true if an int is 32-bit, otherwise 16-bit
  LNG64               = false                           ' Set true if a long is 64-bit, otherwise 32-bit
  CHR16               = true                            ' Set true if a char is 16-bit, otherwise 8-bit

  BIGARRAYS           = false                           ' Set true if allowed arrays larger than 64KB
  
  STACK_SIZE          = 1024                            ' Entries allowed on JVM runtime execution stack

  HEAP_BASE           = $7E00                           ' Address of first free byte above top of heap

CON
  
  INTWORDS            = 1 - INT32                       ' Number of words in an int ( INT32=2, INT16=1 )
  INTBYTES            = INTWORDS * 2                    ' Number of bytes in an int ( INT32=4, INT16=2 )

  LNGWORDS            = ( 1 - LNG64 ) * 2               ' Number of words in a long ( LNG64=4, LNG32=2 )
  LNGBYTES            = LNGWORDS * 2                    ' Number of bytes in a long ( LNG64=8, LNG32=4 )

  CHRWORDS            = ( 1 - CHR16 ) >> 1              ' Number of words in a char ( CHR16=1, CHR8=0  )
  CHRBYTES            = 1 - CHR16                       ' Number of bytes in a char ( CHR16=2, CHR8=1  )

  MODE_SPIN_ONLY      = %010                            ' How the JVM opcode processing can be handled
  MODE_SPIN_THEN_PASM = %011
  MODE_PASM_ONLY      = %100
  MODE_PASM_THEN_SPIN = %110

  
  HEAP_HEAD_TYPE      = 0                               ' Heap chunk Type entry offset
  HEAP_HEAD_BLEN      = 1                               ' Heap chunk ByteLen entry offset
  HEAP_HEAD_DATA      = 2                               ' Heap chunk Data entry offset

  HEAP_WORD           = 1
  HEAP_LONG           = 2
  
VAR

' .-----------------------------------------------------------------------------------------------------.
' |     JVM Runtime Data Structures                                                                     |
' `-----------------------------------------------------------------------------------------------------'

  ' The JVM runtime stack is adjusted for the base
  ' size of int(16) or int(32). Only half as many
  ' longs will be used for int(16) as for int(32).
  ' Keep this as long so it works with INT32.
  
  long stack[ STACK_SIZE >> (INT32+1) ]                 ' Stack for JVM runtime execution

  ' The following ( 'pc' to 'sp' ) must all be longs
  ' and must be contiguous. Add new stack frame items
  ' within this list, not before 'pc' nor after 'sp'
  ' or the stack frames will not work properly.
  
  long pc                                               ' Current JEM bytecode Pc ( absilute hub address )
  long bp                                               ' Current base of stack ( firat param / local )
  long mb                                               ' Method base - first executable in method
  long ct                                               ' Constant table address
  long et                                               ' Exception table address
  long vp                                               ' Variable base ( static fields )
  long sp                                               ' Top of stack pointer ( last pushed )

  long bytecodeStart                                    ' Absolute Hub address of first bytecode
  long jemFileCountAdjust                               ' How to adjust Jem File counts to fit reality

  long heap                                             ' Ptr to last chunk created on heap
  
  long mode                                             ' Type of JVM opcode processing
  long debugging                                        ' Set true if debugging the JVM
  long allowNotImplementedNatives                       ' Set true if unimplemented Native Methods ignored

  long stackAltered                                     ' True if stack has altered flag
  long opcode                                           ' Current opcode
              
PUB Main

  tv.init
  'tv.Start( TV_PIN )
  tv.Str( Version )

  mode := MODE_SPIN_ONLY

  debugging := true
  allowNotImplementedNatives := true
  
  JvmSpinLoop

PRI JvmSpinLoop

' .-----------------------------------------------------------------------------------------------------.
' |     Initialise JVM                                                                                  |
' `-----------------------------------------------------------------------------------------------------'

  sp := @stack - INTBYTES
  bp := sp

  heap := HEAP_BASE
   
  InvokeProgram( jem.BytecodeStart )

' .-----------------------------------------------------------------------------------------------------.
' |     Main JVM Loop                                                                                   |
' `-----------------------------------------------------------------------------------------------------'

  if debugging

    repeat

      ShowStack
      Pause_Ms( 1000 )

      tv.Out( $0D )
      tv.Hex( pc, 4 )
      tv.Out( " " )
      tv.Hex( pc - bytecodeStart, 4 )
      tv.Out( ":" )
      tv.Dec( pc - mb )
      if (pc - mb ) => 0 and ( pc - mb ) < 10
        tv.Out( " " )
      if ( pc - mb ) < 100
        tv.Out( " " )
   
      opcode := byte[ pc++ ]

      tv.Hex( opcode, 2 )
      tv.Out( " " )
      tv.Str( jem.OpcodeNamePtr( opcode ) )

      \JvmSpin

  else
  
    repeat
      opcode := byte[ pc++ ]
      \JvmSpin
      
  if debugging
    ShowStack

  Crashed( String( "JvmSpin terminated" ) )
    
PRI JvmSpin | arrayPtr, arrayLen, arrayIndex, value

  if mode & MODE_PASM_ONLY
    if JvmPasm
       return
    else
      if mode == MODE_PASM_ONLY
        Crashed( String( "PASM did not handle opcode" ) )

  if mode & MODE_SPIN_ONLY
              
    if opcode < jem#Jem_INSTANCEOF
      case opcode

        ' .---------------------------------------------------------------------------------------------.
        ' |     NOP                                                                                 $00 |
        ' |     Meaning: Do nothing.                                                                    |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_NOP:

        ' .---------------------------------------------------------------------------------------------.
        ' |     ICONST_M1                                                                           $01 |
        ' |     Meaning: Push the integer constant -1.                                                  |
        ' |     Description: This bytecode pushes the integer constant -1 onto the stack.               |
        ' |---------------------------------------------------------------------------------------------|
        ' |     ICONST_0                                                                        $02-$07 |
        ' |     Meaning: Push the integer constant <C>.                                                 |
        ' |     Description: This bytecode pushes an integer constant onto the stack. There             |
        ' |     are six associated bytecodes, one for each of the integers 0-5: iconst_0,               |
        ' |     iconst_1, iconst_2, iconst_3, iconst_4, and iconst_5.                                   |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_ICONST_M1, jem#Jem_ICONST_0, jem#Jem_ICONST_1, jem#Jem_ICONST_2, jem#Jem_ICONST_3:

          Push( opcode - jem#Jem_ICONST_0 )

        jem#Jem_ICONST_4, jem#Jem_ICONST_5:

          Push( opcode - jem#Jem_ICONST_4 )

        ' .---------------------------------------------------------------------------------------------.
        ' |     BIPUSH                                                                              $08 |
        ' |     Meaning: Push a 1-byte signed integer.                                                  |
        ' |     Description: The first byte following the opcode, byte1, is interpreted as a            |
        ' |     signed 8-bit value. This value is expanded to a 32-bit integer and pushed onto          |
        ' |     the operand stack.                                                                      |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_BIPUSH:

          Push( Operand_SignedByte )

        ' .---------------------------------------------------------------------------------------------.
        ' |     SIPUSH                                                                              $09 |
        ' |     Meaning: Push a 2-byte signed integer.                                                  |
        ' |     Description: The first and second byte following the opcode, byte1 and byte2,           |
        ' |     are assembled into a signed 16-bit value. This value is expanded to a 32-bit            |
        ' |     integer and pushed onto the operand stack.                                              |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_SIPUSH:

          Push( Operand_SignedWord )

        ' .---------------------------------------------------------------------------------------------.
        ' |     ILOAD                                                                               $0A |
        ' |     Meaning: Load integer from local variable.                                              |
        ' |     Description: The value of the local variable in the current Java frame is               |
        ' |     pushed onto the operand stack. This value must be of type integer.                      |
        ' |---------------------------------------------------------------------------------------------|
        ' |     ILOAD_0                                                                             $0B |
        ' |     ILOAD_1                                                                             $0C |
        ' |     ILOAD_2                                                                             $0D |
        ' |     ILOAD_3                                                                             $0E |
        ' |     Meaning: Load integer from local variable with specific index.                          |
        ' |     Description: The value of the local variable <l> in the current Java frame is           |
        ' |     pushed onto the operand stack.                                                          |
        ' |     This value must be of type integer. There are four of these bytecodes, one for          |
        ' |     each of the integers 0-3: iload_0, iload_1, iload_2, and iload_3.                       |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_ILOAD:

           Push( GetVar( Operand_UnsignedByte ) )

        jem#Jem_ILOAD_0, jem#Jem_ILOAD_1, jem#Jem_ILOAD_2, jem#Jem_ILOAD_3:

           Push( GetVar( opcode - jem#Jem_ILOAD_0 ) )

        ' .---------------------------------------------------------------------------------------------.
        ' |     ISTORE                                                                              $0F |
        ' |     Meaning: Store integer into local variable.                                             |
        ' |     Description: Local variable in the current Java frame is set to a value. This           |
        ' |     value must be an integer.                                                               |
        ' |---------------------------------------------------------------------------------------------|
        ' |     ISTORE_0                                                                            $10 |
        ' |     ISTORE_1                                                                            $11 |
        ' |     ISTORE_2                                                                            $12 |
        ' |     ISTORE_3                                                                            $13 |
        ' |     Meaning: Store integer into local variable with specific index.                         |
        ' |     Description: Local variable <l> in the current Java frame is set to a value.            |
        ' |     This value must be an integer. There are four of these bytecodes, one for each          |
        ' |     of the integers 0-3: istore_0, istore_1, istore_2, and istore_3.                        |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_ISTORE:

           PutVar( Operand_UnsignedByte, Pop )

        jem#Jem_ISTORE_0, jem#Jem_ISTORE_1, jem#Jem_ISTORE_2, jem#Jem_ISTORE_3:

           PutVar( opcode - jem#Jem_ISTORE_0, Pop )

        ' .---------------------------------------------------------------------------------------------.
        ' |     IFxx                                                                                $14 |
        ' |     Meaning: Branch if condition                                                            |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IFEQ: BranchIf( Pop == 0 )
        jem#Jem_IFNE: BranchIf( Pop <> 0 )
        jem#Jem_IFLT: BranchIf( Pop <  0 )
        jem#Jem_IFLE: BranchIf( Pop =< 0 )
        jem#Jem_IFGT: BranchIf( Pop >  0 )
        jem#Jem_IFGE: BranchIf( Pop => 0 )

        ' .---------------------------------------------------------------------------------------------.
        ' |     IF_ICMPxx                                                                           $1A |
        ' |     Meaning: Branch if compare                                                              |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IF_ICMPEQ: BranchIf( PopL == PopR )
        jem#Jem_IF_ICMPNE: BranchIf( PopL <> PopR )
        jem#Jem_IF_ICMPLT: BranchIf( PopL <  PopR )
        jem#Jem_IF_ICMPGE: BranchIf( PopL => PopR )
        jem#Jem_IF_ICMPGT: BranchIf( PopL >  PopR )
        jem#Jem_IF_ICMPLE: BranchIf( PopL =< PopR )

        ' .---------------------------------------------------------------------------------------------.
        ' |     IF_ACMPxx                                                                           $20 |
        ' |     Meaning: Branch if object references equal.                                             |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IF_ACMPEQ: BranchIf( PopObjectRef == PopObjectRef )
        jem#Jem_IF_ACMPNE: BranchIf( PopObjectRef <> PopObjectRef )

        ' .---------------------------------------------------------------------------------------------.
        ' |     IRETURN                                                                             $22 |
        ' |     Meaning: Return integer from function.                                                  |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IRETURN: ReturnFromMethodWithValue
        jem#Jem_RETURN:  ReturnFromMethod

        ' .---------------------------------------------------------------------------------------------.
        ' |     INVOKEVIRTUAL                                                                       $24 |
        ' |     Meaning: Invoke method based on runtime type.                                           |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_INVOKEVIRTUAL:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     INVOKESTATIC                                                                        $25 |
        ' |     Meaning: Invoke a class (static) method.                                                |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_INVOKESTATIC:

          InvokeStatic( Operand_BytecodeAddress )

        ' .---------------------------------------------------------------------------------------------.
        ' |     GETSTATIC                                                                           $26 |
        ' |     Meaning: Get 16bit static field from class.                                             |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_GETSTATIC:

          Push( Peek( vp, Operand_UnsignedWord >> jemFileCountAdjust ) )

        ' .---------------------------------------------------------------------------------------------.
        ' |     PUTSTATIC                                                                           $27 |
        ' |     Meaning: Set 16bit static field in class.                                               |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_PUTSTATIC:

          Poke( vp, Operand_UnsignedWord >> jemFileCountAdjust, Pop )

        ' .---------------------------------------------------------------------------------------------.
        ' |     POP                                                                                 $28 |
        ' |     Meaning: Stack manipulation                                                             |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_POP:     Pop
        jem#Jem_POP2:    Pop2
        jem#Jem_DUP:     Dup
        jem#Jem_DUP_X1:  Dup_X1
        jem#Jem_DUP_X2:  Dup_X2
        jem#Jem_DUP2:    Dup2
        jem#Jem_DUP2_X1: Dup2_X1
        jem#Jem_DUP2_X2: Dup2_X2
        jem#Jem_SWAP:    Swap

        ' .---------------------------------------------------------------------------------------------.
        ' |     IADD, ISUB, INEG, IOR, ISHL, ISHR, IUSHR, IAND, IXOR                                $31 |
        ' |     Meaning: Integer maths                                                                  |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IADD:  Push( PopL +  PopR )
        jem#Jem_ISUB:  Push( PopL -  PopR )
        jem#Jem_INEG:  Push(      -  Pop  )
        jem#Jem_IOR:   Push( PopL |  PopR )
        jem#Jem_ISHL:  Push( PopL << PopR )
        jem#Jem_ISHR:  Push( PopL -> PopR )
        jem#Jem_IUSHR: Push( PopL >> PopR )
        jem#Jem_IAND:  Push( PopL &  PopR )
        jem#Jem_IXOR:  Push( PopL ^  PopR )

        ' .---------------------------------------------------------------------------------------------.
        ' |     I2B                                                                                 $3A |
        ' |     Meaning: Integer to signed byte.                                                        |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_I2B:

          if Tos <> SignExtendByte( Tos )
            Crashed( String( "Signed Int too big for Signed Byte" ) )
             
        ' .---------------------------------------------------------------------------------------------.
        ' |     GOTO                                                                                $3B |
        ' |     Meaning: Branch.                                                                        |
        ' |     Description: Execution proceeds at the signed 16-bit offset from the address            |
        ' |     of this instruction. The offset is constructed from byte1 and byte2.                    |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_GOTO:

          BranchIf( true )

        ' .---------------------------------------------------------------------------------------------.
        ' |     INVOKEINTERFACE                                                                     $3C |
        ' |     Meaning: Invoke interface method.                                                       |
        ' |     Note                                                                                    |
        ' |     Unlike invokevirtual and invokespecial, the method block does not indicate the          |
        ' |     number of available arguments. This number is taken from the bytecode.                  |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_INVOKEINTERFACE:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     ARRAYLENGTH                                                                         $3D |
        ' |     Meaning: Get length of array.                                                           |
        ' |     Description: The length of the array is determined and replaces aref on the             |
        ' |     top of the stack. The aref must be a reference to an array object.                      |
        ' |     Note                                                                                    |
        ' |     A NullPointerException is thrown if the aref is null.                                   |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_ARRAYLENGTH:

          PopArrayRef( jem#T_ANY, @arrayPtr, @arrayLen )
          Push( arrayLen)
          
        ' .---------------------------------------------------------------------------------------------.
        ' |     ATHROW                                                                              $3E |
        ' |     Meaning: Throw exception or error.                                                      |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_ATHROW:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     CHECKCAST                                                                           $3F |
        ' |     Meaning: Make sure object is of given type.                                             |
        ' |     Description: Determines whether the referenced object can be cast to be a               |
        ' |     reference to an object of another class. A null object reference can be cast to         |
        ' |     any class. Otherwise, the referenced object must be an instance of the expected         |
        ' |     class or one of its superclasses. byte1 and byte2 are used to construct an              |
        ' |     index into the constant pool of the current class, which is presumed to be a            |
        ' |     class name that can be resolved to a class pointer.                                     |
        ' |     Note                                                                                    |
        ' |     A ClassCastException is thrown if the referenced object cannot be cast to the           |
        ' |     expected class.                                                                         |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_CHECKCAST:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     Catch all unimplemented opcodes                                                         |
        ' `---------------------------------------------------------------------------------------------'

        other: NotImplemented

    else
      case opcode

        ' .---------------------------------------------------------------------------------------------.
        ' |     INSTANCEOF                                                                          $40 |
        ' |     Meaning: Determine if an object is of given type and return result.                     |
        ' |     Description: Determines whether the referenced object can be cast to be a               |
        ' |     reference to an object of the expected class. This instruction will overwrite           |
        ' |     the object reference with 1 if the object is an instance of the expected class          |
        ' |     or one of its superclasses. Otherwise, the object reference is overwritten by 0.        |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_INSTANCEOF:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     WIDE                                                                                $41 |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_WIDE:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     IINC                                                                                $42 |
        ' |     Meaning: Increment local variable by constant.                                          |
        ' |     Description: Local variable at byte1 in the current Java frame must contain an          |
        ' |     integer. Its value is incremented by the value byte2, where byte2 is treated as         |
        ' |     a signed 8-bit quantity.                                                                |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IINC:

          IncVar( Operand_UnSignedByte, Operand_SignedByte )

        ' .---------------------------------------------------------------------------------------------.
        ' |     GETFIELD                                                                            $43 |
        ' |     Meaning: Get 16bit field from object.                                                   |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_GETFIELD:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     PUTFIELD                                                                            $44 |
        ' |     Meaning: Set 16bit field in object.                                                     |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_PUTFIELD:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     NEW                                                                                 $45 |
        ' |     Meaning: Create new object.                                                             |
        ' |     Description: byte1 and byte2 are used to construct an index into the constant           |
        ' |     pool of the current class, which must be a class name that can be resolved to a         |
        ' |     class pointer. A new instance of that class is then created, and a reference to         |
        ' |     the object is pushed on the stack.                                                      |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_NEW:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     NEWARRAY                                                                            $46 |
        ' |     Meaning: Allocate new array.                                                            |
        ' |     Description: A new array of a specific array type, capable of holding size              |
        ' |     elements, is allocated. The result is a reference to the new object. Allocation         |
        ' |     of an array large enough to contain size items of the specific array type is            |
        ' |     attempted and all elements of the array are initialized to 0.                           |
        ' |     size represents the number of elements in the new array and must be an                  |
        ' |     integer. The result is stored with an internal code that indicates the type of          |
        ' |     array to allocate. Possible values for the type of array are as follows:                |
        ' |     T_BOOLEAN(4), T_chAR(5), T_FLOAT(6), T_DOUBLE(7), T_BYTE(8), T_SHORT(9),                |
        ' |     T_INT(10), and T_LONG(11).                                                              |
        ' |     Note                                                                                    |
        ' |     A NegativeArraySizeException is thrown if size is less than 0. An                       |
        ' |     OutOfMemoryError is thrown if there is not enough memory to allocate the array.         |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_NEWARRAY:

          arrayLen := Pop

          Operand_UnsignedByte ' ***** HACK **** This is $00 so what's it mean ?

          Push( NewArrayRef( jem#T_INT, arrayLen ) )

        ' .---------------------------------------------------------------------------------------------.
        ' |     ANEWARRAY                                                                           $47 |
        ' |     Meaning: Allocate new array of objects.                                                 |
        ' |     Description: A new array of the indicated class type and capable of holding             |
        ' |     size elements is allocated. The result is a reference to the new object.                |
        ' |     Allocation of an array large enough to contain size elements of the given class         |
        ' |     type is attempted and all elements of the array are initialized to null.                |
        ' |     size represents the number of elements in the new array and must be an                  |
        ' |     integer. byte1 and byte2 are used to construct an index into the constant pool          |
        ' |     of the current class. When the item at that index is resolved, the resulting            |
        ' |     entry must be a class.                                                                  |
        ' |     The anewarray instruction is used to create a single-dimension array.                   |
        ' |     Note                                                                                    |
        ' |     A NegativeArraySizeException is thrown if size is less than 0. An                       |
        ' |     OutOfMemoryError is thrown if there is not enough memory to allocate the array.         |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_ANEWARRAY:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     IALOAD                                                                              $48 |
        ' |     Meaning: Load integer from array.                                                       |
        ' |     Description: The integer value at the array index is retrieved and pushed onto          |
        ' |     the top of the stack. The aref must be a reference to an array of integers;             |
        ' |     likewise, the index into the array must be an integer.                                  |
        ' |     Note                                                                                    |
        ' |     A NullPointerException is thrown if aref is null. An                                    |
        ' |     ArrayIndexOutOfBoundsException is thrown if the array index is not within the           |
        ' |     bounds of the array.                                                                    |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IALOAD:

          arrayIndex := Pop
          PopArrayRef( jem#T_INT, @arrayPtr, @arrayLen )

          if arrayIndex < 0 or arrayIndex => arrayLen
            Crashed( String( "Array Index out of bounds" ) )

          if INT32
            push( long[ arrayPtr + ( arrayIndex << 2 ) ] )
          else
            Push( SignExtendWord( word[ arrayPtr + ( arrayIndex << 1 ) ] ) )

        ' .---------------------------------------------------------------------------------------------.
        ' |     IASTORE                                                                             $49 |
        ' |     Meaning: Store into integer array.                                                      |
        ' |     Description: An integer value is popped off the stack and stored in the array           |
        ' |     at the index. The aref must be a reference to an array of integer values, and           |
        ' |     the index into the array must be an integer as well.                                    |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IASTORE:

          value      := Pop
          arrayIndex := Pop
          PopArrayRef( jem#T_INT, @arrayPtr, @arrayLen )

          if arrayIndex < 0 or arrayIndex => arrayLen
            Crashed( String( "Array Index out of bounds" ) )
             
          if INT32
            long[ arrayPtr + ( arrayIndex << 2 ) ] := value
          else
            word[ arrayPtr + ( arrayIndex << 1 ) ] := value

        ' .---------------------------------------------------------------------------------------------.
        ' |     BALOAD                                                                              $4A |
        ' |     Meaning: Load signed byte from array.                                                   |
        ' |     Description: The signed byte value at the array index is retrieved, expanded            |
        ' |     to an integer, and pushed onto the top of the stack. The aref must be a                 |
        ' |     reference to an array of signed byte values, and the index into the array must          |
        ' |     be an integer.                                                                          |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_BALOAD:

          arrayIndex := Pop
          PopArrayRef( jem#T_BYTE, @arrayPtr, @arrayLen )

          if arrayIndex < 0 or arrayIndex => arrayLen
            Crashed( String( "Array Index out of bounds" ) )

          Push( SignExtendByte( byte[ arrayPtr + arrayIndex ] ) )

        ' .---------------------------------------------------------------------------------------------.
        ' |     BASTORE                                                                             $4B |
        ' |     Meaning: Store into signed byte array.                                                  |
        ' |     Description: An integer is popped off the stack, converted to a signed byte,            |
        ' |     and stored in the array at the index. The arrayref must be a reference to an            |
        ' |     array of signed bytes, and the index into the array must be an integer. If the          |
        ' |     integer value is too large to be a signed byte, it is truncated.                        |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_BASTORE:

          value      := Pop
          arrayIndex := Pop
          PopArrayRef( jem#T_BYTE, @arrayPtr, @arrayLen )
          
          if arrayIndex < 0 or arrayIndex => arrayLen
            Crashed( String( "Array Index out of bounds" ) )

          if value <> SignExtendByte( value )
            Crashed( String( "Signed Int too big for Signed Byte" ) )
             
          byte[ arrayPtr + arrayIndex ] := value

        ' .---------------------------------------------------------------------------------------------.
        ' |     LDC                                                                                 $4C |
        ' |     Meaning: Push item from constant pool.                                                  |
        ' |     Description: byte1 is used as an unsigned 8-bit index into the constant pool            |
        ' |     of the current class. After the item at that index is resolved, it is pushed            |
        ' |     onto the stack.                                                                         |
        ' |     Note                                                                                    |
        ' |     This bytecode instruction throws an OutOfMemoryError if a String object is              |
        ' |     being pushed and there is not enough memory to allocate space for it.                   |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_LDC:

          LoadConstant( 0, Operand_UnsignedByte )

        ' .---------------------------------------------------------------------------------------------.
        ' |     LDC_W                                                                               $4D |
        ' |     Meaning: Push 2-byte item from constant pool.                                           |
        ' |     Description: byte1 and byte2 are used to construct an unsigned 16-bit index             |
        ' |     into the constant pool of the current class. After the item at that index is            |
        ' |     resolved, it is pushed onto the stack.                                                  |
        ' |     Note                                                                                    |
        ' |     This bytecode instruction throws an OutOfMemoryError if a String object is              |
        ' |     being pushed and there is not enough memory to allocate space for it.                   |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_LDC_W:

          LoadConstant( 2, Operand_UnsignedWord )

        ' .---------------------------------------------------------------------------------------------.
        ' |     LDC2_W                                                                              $4E |
        ' |     Meaning: Push long or double from constant pool.                                        |
        ' |     Description: byte1 and byte2 are used to construct an unsigned 16-bit index             |
        ' |     into the constant pool of the current class. After the two-word constant at             |
        ' |     that index is resolved, it is pushed onto the stack.                                    |
        ' |     Note                                                                                    |
        ' |     This bytecode instruction throws an OutOfMemoryError if a String object is              |
        ' |     being pushed and there is not enough memory to allocate space for it.                   |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_LDC2_W:

          LoadConstant( 4, Operand_UnsignedWord )

        ' .---------------------------------------------------------------------------------------------.
        ' |     IMUL, IDIV, IREM                                                                    $4F |
        ' |     Meaning: Integer maths                                                                  |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IMUL: Push( PopL *  PopR )
        jem#Jem_IDIV: Push( PopL /  PopR )
        jem#Jem_IREM: Push( PopL // PopR )

        ' .---------------------------------------------------------------------------------------------.
        ' |     JSR                                                                                 $52 |
        ' |     Meaning: Jump subroutine.                                                               |
        ' |     Description: Execution proceeds at the signed 16-bit offset from the address            |
        ' |     of this instruction. The offset is constructed from byte1 and byte2. The                |
        ' |     address of the instruction immediately following the current instruction is             |
        ' |     pushed onto the stack to provide a return address from the subroutine.                  |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_JSR:

          Push( pc )
          BranchIf( true )

        ' .---------------------------------------------------------------------------------------------.
        ' |     RET                                                                                 $53 |
        ' |     Meaning: Return from subroutine.                                                        |
        ' |     Description: Local variable in the current Java frame must contain a return             |
        ' |     address. The contents of the local variable are written into the program                |
        ' |     counter.                                                                                |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_RET:

          pc := GetVar( Operand_UnsignedByte )

        ' .---------------------------------------------------------------------------------------------.
        ' |     TABLESWITCH                                                                         $54 |
        ' |     Meaning: Access jump table by index and jump.                                           |
        ' |     Description: Immediately after the tableswitch instruction, padding consisting          |
        ' |     of 0-3 zeros is inserted so that the next byte begins at an address that is a           |
        ' |     multiple of four. A series of offsets follows the padding. These offsets are            |
        ' |     signed 4-byte quantities and consist of a default offset, a low offset, and a           |
        ' |     high offset, followed by additional high - low + 1 offsets. These additional            |
        ' |     offsets are treated as a 0-based jump table.                                            |
        ' |     The index the offsets point to must be an integer. If the index is less than            |
        ' |     the low offset or greater than the high offset, the default offset is added to          |
        ' |     the address of the current instruction. Otherwise, the low offset is subtracted         |
        ' |     from the index, and the element at the position index - low offset in the jump          |
        ' |     table is extracted and added to the address of the current instruction.                 |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_TABLESWITCH:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     LOOKUPSWITCH                                                                        $55 |
        ' |     Meaning: Access jump table by key match and jump.                                       |
        ' |     Description: Immediately after the lookupswitch instruction, padding                    |
        ' |     consisting of 0-3 zeros is                                                              |
        ' |     inserted so that the next byte begins at an address that is a multiple of               |
        ' |     four. A series of pairs of offsets follows the padding. These offsets are               |
        ' |     signed 4-byte quantities. The first item in the pair is the default offset, and         |
        ' |     the second item in the pair gives the number of pairs that follow. The                  |
        ' |     additional pairs consist of a match and an offset.                                      |
        ' |     The integer key on the stack is then compared against each of the matches. If           |
        ' |     the key is equal to one of the matches, the offset is added to the address of           |
        ' |     the current instruction. If the key does not match any of the matches, the              |
        ' |     default offset is added to the address of the current instruction.                      |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_LOOKUPSWITCH:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     Catch all unimplemented opcodes                                                         |
        ' `---------------------------------------------------------------------------------------------'

        other: NotImplemented

' .-----------------------------------------------------------------------------------------------------.
' |     Native Method Handling                                                                          |
' `-----------------------------------------------------------------------------------------------------'

PRI NativeMethod( nativeMethodNumber, numberOfParameters ) | i

  ' On entry the stack will hold the numberOfparameters needed to execute the native method, with BP 
  ' set as pointing to the first parameter, SP pointing to the last parameter. Pop() will remove in
  ' reverse order from SP down to BP. The stack can be pushed and popped at will during the Native
  ' Method handler so long as it is not overflowed or underflowed.
  '
  ' When leaving the Native Method handler any return result should have been pushed to the stack and
  ' a "result := true" issued to indicate there is a value returned, otherwise 'result' should not be
  ' set and the Native Method handler simply exited.
  '
  ' For example a native method call to "stamp.core.CPU.Count(A,B,C)" will deliver a numberOfParameters
  ' count of 3 and a stack as follows -
  ' 
  '       .---. --.         
  ' BP -->| A |   |
  '       | B |   |-- Number of Parameters = 3
  ' SP -->| C |   |
  '       `---' --'

  opcode := -1

  if debugging
    tv.Str( String( $0D,"Native       " ) )
    tv.Hex( nativeMethodNumber, 2 )
    tv.Out( " " )
    tv.Str( jem.nativeMethodNamePtr( nativeMethodNumber ) )
    tv.Out( "(" )
    if numberOfParameters
      repeat i from 0 to numberOfParameters-1
        if i <> 0
          tv.Out( "," )
        tv.Dec( GetVar( i ) )
    tv.Out( ")" )

  case nativeMethodNumber

    jem#Native_CPU_writeRegister:

      NotImplemented

    jem#Native_CPU_readRegister:

      NotImplemented
      Push(0)
      result := true
      
    jem#Native_CPU_writeObject:

      NotImplemented

    jem#Native_CPU_delay:

      NotImplemented

    jem#Native_CPU_message:

      NotImplemented

    jem#Native_CPU_count:

      NotImplemented
      Push(0)
      result := true

    jem#Native_CPU_rcTime:

      NotImplemented
      Push(0)
      result := true

    jem#Native_CPU_pulseIn:

      NotImplemented
      Push(0)
      result := true

    jem#Native_CPU_pulseOut:

      NotImplemented

    jem#Native_CPU_readPin:

      NotImplemented
      Push(0)
      result := true

    jem#Native_CPU_writePin:

      NotImplemented

    jem#Native_CPU_shiftIn:

      NotImplemented
      Push(0)
      result := true

    jem#Native_CPU_shiftOut:

      NotImplemented

    jem#Native_CPU_carry:

      NotImplemented
      Push(0)
      result := true

    jem#Native_CPU_uninstall:

      NotImplemented

    jem#Native_CPU_install:

      NotImplemented

    jem#Native_CPU_nap:

      NotImplemented

    jem#Native_CPU_setOutput:

      NotImplemented

    jem#Native_CPU_setInput:

      NotImplemented

    jem#Native_CPU_readSRAM:

      NotImplemented
      Push(0)
      result := true

    jem#Native_CPU_writeSRAM:

      NotImplemented

    jem#Native_EEPROM_eeRead:

      NotImplemented
      Push(0)
      result := true

    jem#Native_EEPROM_eeWrite:

      NotImplemented

    jem#Native_Uart_txInit:

      NotImplemented

    jem#Native_Uart_rxInit:

      NotImplemented

    jem#Native_Uart_rxRead:

      NotImplemented
      Push(0)
      result := true

    jem#Native_Callback_returnTwo:

      NotImplemented
      Push(0)
      result := true

    jem#Native_Timer_init:

      NotImplemented

    jem#Native_Timer_latch:

      NotImplemented

    jem#Native_PWM_install:

      NotImplemented

    jem#Native_PWM_updateInternal:

      NotImplemented

    jem#Native_DAC_install:

      NotImplemented

    jem#Native_ADC_install:

      NotImplemented

    jem#Native_Terminal_getByte:

      NotImplemented
      Push(0)
      result := true

    other:
    
      NotImplemented

' .-----------------------------------------------------------------------------------------------------.
' |     Deal with unimplemened opcodes and forced PASM calls                                            |
' `-----------------------------------------------------------------------------------------------------'

PRI NotImplemented

  if opcode < 0
    if not allowNotImplementedNatives
      Crashed( String( "Native Method not implemented" ) )
  else
    if mode == MODE_SPIN_ONLY
      Crashed( String( "Spin did not handle opcode" ) )
    else
      if mode == MODE_SPIN_THEN_PASM
        if JvmPasm
          abort
        else
          Crashed( String( "Spin then PASM did not handle opcode" ) )
      else
        Crashed( String( "PASM then Spin did not handle opcode" ) )

PRI UsePasm

  if JvmPasm
    abort
  else
    Crashed( String( "Forced PASM did not handle opcode" ) )

' .-----------------------------------------------------------------------------------------------------.
' |     Bytecode Operand Handling                                                                       |
' `-----------------------------------------------------------------------------------------------------'

PRI Operand_UnSignedByte : operand

  operand := byte[ pc++ ]

  if debugging
    tv.Out( " " )
    tv.Out( "$" )
    tv.Hex( operand, 2 )
  
PRI Operand_SignedByte : operand

  operand := SignExtendByte( Operand_UnSignedByte )

PRI Operand_UnSignedWord : operand

  operand := byte[ pc++ ] << 8
  operand |= byte[ pc++ ]

  if debugging
    tv.Out( " " )
    tv.Out( "$" )
    tv.Hex( operand,4)
  
PRI Operand_SignedWord : operand

  operand := SignExtendWord( Operand_UnSignedWord )

PRI Operand_BytecodeAddress : operand

  operand := Operand_UnSignedWord + bytecodeStart

PRI SignExtendByte( n )

  result := ~ n

PRI SignExtendWord( n )

  result := ~~ n

' .-----------------------------------------------------------------------------------------------------.
' |     Stack Pushing                                                                                   |
' `-----------------------------------------------------------------------------------------------------'

PRI Push( n )

  if INT32
    PushLong( n )
  else
    PushWord( n )
    
PRI PushLong( n )

  stackAltered := true
  if INT32
    sp += 4
    if sp > @stack + ( STACK_SIZE << INTWORDS )
      Crashed( String( "Stack overflow" ) )
    long[ sp ] := n
  else
    Crashed( String( "Cannot use PushLong if not INT32" ) )
   
PRI PushWord( n )

  stackAltered := true
  sp += 2
  if sp > @stack + ( STACK_SIZE << INTWORDS )
    Crashed( String( "Stack overflow" ) )
  word[ sp ] := n

PRI PushAddr( n )

  if INT32
    PushLong( n )
  else
    if ChipVer == 1
      PushWord( n )
    else
      PushWord( n >> 16 )
      PushWord( n       )
      
' .-----------------------------------------------------------------------------------------------------.
' |     Stack Popping                                                                                   |
' `-----------------------------------------------------------------------------------------------------'

PRI Pop

  if INT32
    result := PopLong
  else
    result := PopWord

PRI PopLong

  stackAltered := true
  if INT32
    if sp < @stack
      Crashed( String( "Stack underflow" ) )
    result := long[ sp ]
    sp -= 4
  else
    Crashed( String( "Cannot use PopLong if not INT32" ) )
    
PRI PopWord

  stackAltered := true
  if sp < @stack
    Crashed( String( "Stack underflow" ) )
  result := SignExtendWord( word[ sp ] )
  sp -= 2

PRI PopAddr

  if INT32
    result := PopLong
  else
    if ChipVer == 1
      result := PopWord
    else
      result := PopWord
      result |= PopWord << 16

PRI Tos

  result := Pop
  Push( result )
        
' .-----------------------------------------------------------------------------------------------------.
' |     Stack Manipulation                                                                              |
' `-----------------------------------------------------------------------------------------------------'

PRI PopL : lhs | rhs

  '       .---.           .---.
  '       | L |     SP -->| R |
  '       |---|           `---'
  ' SP -->| R |
  '       `---'

  rhs := Pop
  lhs := Pop
  Push( rhs )

PRI PopR : rhs

  rhs := Pop

PRI Pop2

  Pop
  Pop

PRI Dup | a

  '       .---.           .---.
  ' SP -->| A |           | A |
  '       `---'           |---|
  '                 SP -->| A |
  '                       `---'

  a := Pop

  Push( a )
  Push( a )

PRI Dup_X1 | a, b

  '       .---.           .---.
  '       | B |           | A |
  ' SP -->| A |           |---|
  '       `---'           | B |
  '                 SP -->| A |
  '                       `---'
  a := Pop
  b := Pop

  Push( a )
  Push( b )
  Push( a )

PRI Dup_X2 | a, b, c

  '       .---.           .---.
  '       | C |           | A |
  '       | B |           |---|
  ' SP -->| A |           | C |
  '       `---'           | B |
  '                 SP -->| A |
  '                       `---'
  a := Pop
  b := Pop
  c := Pop

  Push( a )
  Push( c )
  Push( b )
  Push( a )

PRI Dup2 | a, b

  '       .---.           .---.
  '       | B |           | B |
  ' SP -->| A |           | A |
  '       `---'           |---|
  '                       | B |
  '                 SP -->| A |
  '                       `---'

  a := Pop
  b := Pop

  Push( b )
  Push( a )
  Push( b )
  Push( a )

PRI Dup2_X1 | a, b, c

'       .---.           .---.
'       | C |           | B |
'       | B |           | A |
' SP -->| A |           |---|
'       `---'           | C |
'                       | B |
'                 SP -->| A |
'                       `---'

  a := Pop
  b := Pop
  c := Pop

  Push( b )
  Push( a )
  Push( c )
  Push( b )
  Push( a )

PRI Dup2_X2 | a, b, c, d

  '       .---.           .---.
  '       | D |           | B |
  '       | C |           | A |
  '       | B |           |---|
  ' SP -->| A |           | D |
  '       `---'           | C |
  '                       | B |
  '                 SP -->| A |
  '                       `---'

  a := Pop
  b := Pop
  c := Pop
  d := Pop

  Push( b )
  Push( a )
  Push( d )
  Push( c )
  Push( b )
  Push( a )

PRI Swap | a, b

  '       .---.           .---.
  '       | B |           | A |
  ' SP -->| A |     SP -->| B |
  '       `---'           `---'

  a := Pop
  b := Pop

  Push( a )
  Push( b )

' .-----------------------------------------------------------------------------------------------------.
' |    Variable Handling                                                                                |
' `-----------------------------------------------------------------------------------------------------'

PRI GetVar( offset )

  result := Peek( bp, offset )

PRI PutVar( offset, n )

  Poke( bp, offset, n )

PRI IncVar( offset, n )

  PutVar( offset, GetVar( offset ) + n )
  
PRI Peek( base, offset )

  if INT32
    result := long[ base + ( offset << 2 ) ]
  else
    result := SignExtendWord( word[ base + ( offset << 1 ) ] )

PRI Poke( base, offset, n )

  if INT32
    long[ base + ( offset << 2 ) ] := n
  else
    word[ base + ( offset << 1 ) ] := n

' .-----------------------------------------------------------------------------------------------------.
' |    Arrays                                                                                           |
' `-----------------------------------------------------------------------------------------------------'
'
' How an array is held in memory
' .------.--------.---.---.- - -.---.
' | Type | BytLen | 0 | 1 |     | N |
' `------^--------^---^---^- - -^---'
'
' An Array Reference is a pointer to the Type
' The Type is an INT that holds a value of jem#T_XXXX
' The BytLen is the number of bytes used to hold the array data
' A returned Array Start is a pointer to the first data item
' The Type and Bytlen are always Int(16) or Int(32) depending upon ChipVer *NOT* INT32 
' The array data entity size depends upon Typ

PRI NewArrayRef( typ, len )

  result := GetFromHeap( typ, len )
    
PRI PopArrayRef( typ, atPtr, atLen ) | arrayTyp, bytLen, datPtr

  result := Pop

  if result == 0
    Crashed( String( "Got Null Pointer for Array Ref" ) )
  else
     
    GetHeapInfo( result, @arrayTyp, @bytLen, @datPtr )
    long[ atPtr ] := datPtr
    long[ atLen ] := bytLen >> BytesShiftPerType( arrayTyp )

    if arrayTyp <> typ
      if typ <> jem#T_ANY
        Crashed( String( "Incorrect Array Type" ) )
     
' .-----------------------------------------------------------------------------------------------------.
' |    Objects                                                                                          |
' `-----------------------------------------------------------------------------------------------------'

PRI PopObjectRef

  NotImplemented
  
' .-----------------------------------------------------------------------------------------------------.
' |    Heap Management                                                                                  |
' `-----------------------------------------------------------------------------------------------------'
'
' How an array is held in memory
' .------.--------.---.---.- - -.---.
' | Type | BytLen | 0 | 1 |     | N |
' `------^--------^---^---^- - -^---'
'
' An Array Reference is a pointer to the Type
' The Type is an INT that holds a value of jem#T_XXXX
' The BytLen is the number of bytes used to hold the array data
' A returned Array Start is a pointer to the first byte of data
' The Type and Bytlen are always Int(16) or Int(32) depending upon INT32
' The array data entity size depends upon Typ

PRI GetFromHeap( typ, len ) | bytLen, i

  bytLen := len << BytesShiftPerType(typ)

  ' tv.Out($0D)
  ' tv.Str( String( "Array Typ  " ) )
  ' tv.Dec( typ )
  ' tv.Out( " ")
  ' tv.Str( jem.TypeNamePtr(typ) )
  ' tv.Out($0D)
  ' tv.Str( String( "Array Len  " ) )
  ' tv.Dec( len )
  ' tv.Out($0D)
  ' tv.Str( String( "Array BytLen " ) )
  ' tv.Dec( bytLen )
  ' Pause_Ms( 10_000 )
  
  if ( ChipVer == 1 ) or ( not BIGARRAYS )
    heap &= $FFFF_FFFE
    heap -= bytLen + ( HEAP_HEAD_DATA << HEAP_WORD )
    word[ heap + ( HEAP_HEAD_TYPE << HEAP_WORD ) ] := typ
    word[ heap + ( HEAP_HEAD_BLEN << HEAP_WORD ) ] := bytLen
    i  := heap + ( HEAP_HEAD_DATA << HEAP_WORD )
  else
    heap &= $FFFF_FFFC
    heap -= bytLen + ( HEAP_HEAD_DATA << HEAP_LONG )
    long[ heap + ( HEAP_HEAD_TYPE << HEAP_LONG ) ] := typ
    long[ heap + ( HEAP_HEAD_BLEN << HEAP_LONG ) ] := bytLen
    i  := heap + ( HEAP_HEAD_DATA << HEAP_LONG )

  repeat bytLen
    byte[ i++ ] := 0
    
  return heap

PRI GetHeapInfo( ptr, atTyp, atLen, atDat ) : nextPtr | typ, Bytlen, datPtr
 
  if ( ChipVer == 1 ) or ( not BIGARRAYS )
    typ    := word[ ptr + ( HEAP_HEAD_TYPE << HEAP_WORD ) ]
    bytLen := word[ ptr + ( HEAP_HEAD_BLEN << HEAP_WORD ) ]
    datPtr := ptr + ( HEAP_HEAD_DATA << HEAP_WORD )
    nextPtr := datPtr + bytLen
    if nextPtr & 1
      nextPtr++
  else
    typ    := long[ ptr + ( HEAP_HEAD_TYPE << HEAP_LONG ) ]
    bytLen := long[ ptr + ( HEAP_HEAD_BLEN << HEAP_LONG ) ]
    datPtr := ptr + ( HEAP_HEAD_DATA << HEAP_LONG )
    nextPtr := datPtr + bytLen
    if nextPtr & %11
      nextPtr := ( nextPtr | %11 ) + 1

  if atTyp
    long[ atTyp ] := typ
  if atLen
    long[ atLen ] := bytLen
  if atdat
    long[ atDat ] := datPtr
      
PRI BytesShiftPerType( typ )

  case typ & jem#T_ANY
  
    jem#T_BOOLEAN: result := ( >| 1        ) - 1
    jem#T_CHAR:    result := ( >| CHRBYTES ) - 1
    jem#T_FLOAT:   result := ( >| 2        ) - 1
    jem#T_DOUBLE:  result := ( >| 2        ) - 1
    jem#T_BYTE:    result := ( >| 1        ) - 1
    jem#T_SHORT:   result := ( >| 2        ) - 1
    jem#T_INT:     result := ( >| INTBYTES ) - 1
    jem#T_LONG:    result := ( >| LNGBYTES ) - 1
    jem#T_ANY:     result := ( >| INTBYTES ) - 1
    
    other:     Crashed( String( "Unknown Type" ) )

' .-----------------------------------------------------------------------------------------------------.
' |    Constant Handling                                                                                |
' `-----------------------------------------------------------------------------------------------------'

PRI LoadConstant( constantSizeInBytes, tableIndex ) | constAdr

  '              ; String index table.
  ' 0B25: dw $0B ; String offset hi = $0B
  ' 0B26: dw $2B ;               lo = $2B
  ' 0B27: dw $0B ; String offset hi = $0B
  ' 0B28: dw $35 ;               lo = $35
  ' 0B29: dw $0B ; String offset hi = $0B
  ' 0B2A: dw $43 ;               lo = $43

  constAdr := byte[ ct + tableIndex << 1 + 0 ] << 8
  constAdr |= byte[ ct + tableIndex << 1 + 1 ] 
  constAdr += bytecodeStart 
  
  ' 0B35: dw $01 ; String class  hi = $01
  ' 0B36: dw $C9 ;               lo = $C9
  ' 0B37: dw $00 ; String length hi = $00
  ' 0B38: dw $04 ;               lo = $04
  ' 0B39: dw $0B ; Pointer       hi = $0B
  ' 0B3A: dw $3B ;               lo = $3B
  ' 0B3B: dw $01 ; Object class  hi = $01
  ' 0B3C: dw $20 ;               lo = $20
  ' 0B3D: dw $00 ; String length hi = $00
  ' 0B3E: dw $04 ;               lo = $04
  ' 0B3F: dw "n" ; String Data   "null"
  ' 0B40: dw "u"
  ' 0B41: dw "l"
  ' 0B42: dw "l"

  mode := 0 ' We messed with stuff so don't let PASM try !
  NotImplemented

' .-----------------------------------------------------------------------------------------------------.
' |    Flow Control                                                                                     |
' `-----------------------------------------------------------------------------------------------------'

PRI BranchIf ( condition ) | offset

  offset := Operand_SignedWord - 3

  if debugging
    tv.Out( " " )
    tv.Out( "(" )
    tv.Out( " " )
    tv.Dec( pc + offset - mb )
    tv.Out( " " )
    tv.Out( ")" )
  
  if condition
    pc += offset

PRI InvokeProgram( bytecodeStartPtr ) | numberOfFields, mainMethodAdr, constantTable

  bytecodeStart := bytecodeStartPtr

  ' 0000: dw $A4 ; Placeholder for run/debug.
  ' 0001: dw $0  ; Placeholder for the length.
  ' 0002: dw $01 ; Version: 0.1
  ' 0003: dw 16  ; Number of static fields = 8
  '
  ' 0004: dw $00 ; `main' method: Offset hi = $00
  ' 0005: dw $1E ;                       lo = $1E
  '
  ' 0006: dw $01 ; java.lang.Object class: Offset hi = $01
  ' 0007: dw $12 ;                                lo = $12
  '
  ' 0008: dw $01 ; java.lang.String class: Offset hi = $01
  ' 0009: dw $BB ;                                lo = $BB
  '
  ' 000A: dw $00 ; Throwable.throwVMException method: Offset hi = $00
  ' 000B: dw $36 ;                                           lo = $36
  '
  ' 000C: dw $0B ; String table offset hi = $0B
  ' 000D: dw $03 ;                     lo = $03
  '
  ' 000E: dw $00 ; Callback method offset hi = $00
  ' 000F: dw $00 ;                        lo = $00
  '
  ' 0010: dw $01 ; java.lang.OutOfMemoryError class: Offset hi = $01
  ' 0011: dw $95 ;                                          lo = $95
  '
  ' 0012: dw $0B ; Stack base hi = $0B
  ' 0013: dw $40 ;            lo = $40

  case byte[ bytecodeStart + $0002 ] ' File Version

    ' Standard Javelin Int(16), no Long
    $01   : if INT32
              tv.Str( String( $0D,"Jem File INT16's auto-promoted to INT32" ) )
            jemFileCountAdjust := 1
            
'== ' Int(16), Long(32)
'== $80   :  if INT32
'==           tv.Str( String( $0D,"Jem File INT16's auto-promoted to INT32" ) )
    
'== ' Int(32), Long(32)
'== $C0   : if not INT32
'==           Crashed( String( "Jem File requires INT32" ) )
            
    other : Crashed( String( "Unknown Jem File Version" ) )

  numberOfFields := byte[ bytecodeStart + $0003 ] >> jemFileCountAdjust
   
  mainMethodAdr  := byte[ bytecodeStart + $0004 ] << 8
  mainMethodAdr  |= byte[ bytecodeStart + $0005 ] 
  
  constantTable  := byte[ bytecodeStart + $000C ] << 8
  constantTable  |= byte[ bytecodeStart + $000D ] 

  ' Create a chunk which we can hold our static fields in, and point to the fields ...

  if ( ChipVer == 1 ) or ( not BIGARRAYS )
    vp := GetFromHeap( jem#T_INT, numberOfFields) + ( HEAP_HEAD_DATA << HEAP_WORD )
  else
    vp := GetFromHeap( jem#T_INT, numberOfFields) + ( HEAP_HEAD_DATA << HEAP_LONG )
  
  ct := constantTable + bytecodeStart
  
  CallMethod( mainMethodAdr + bytecodeStart, true )
  
PRI InvokeStatic( methodBlockAdr )

  CallMethod( methodBlockAdr, false )

PRI CallClass( classBlockAdr ) | superClassAdr, numberOfFields 

  '              ; Class = `java/lang/String'
  ' 01BB: dw $01 ; 0 - Superclass offset hi = $01
  ' 01BC: dw $12 ; 1 - Superclass offset lo = $12
  ' 01BD: dw 4   ; 2 - Number of (non-final, non-static) fields = 2

  superClassAdr  := byte[ classBlockAdr + 0 ] << 8
  superClassAdr  |= byte[ classBlockAdr + 1 ] 
  numberOfFields := byte[ classBlockAdr + 2 ] >> jemFileCountAdjust

  Crashed( String( "CallClass not written yet" ) )
  
PRI CallMethod( methodBlockAdr, hasClinit ) | frameStackSize, numberOfParameters, numberOfLocals, flags, methodAdr, exceptionTable, i

  '              ; Method = 'MyMethod'
  ' 004E: dw 4   ; 0 - Number of parameters = 2
  ' 004F: dw 6   ; 1 - Number of locals = 3
  ' 0050: dw $00 ; 2 - Flags     = $00
  ' 0051: dw $04 ; 3 - Offset hi = $04
  ' 0052: dw $E8 ; 4 - Offset lo = $E8
  ' 0053: dw $00 ; 5 - Empty exception table hi = $00
  ' 0054: dw $00 ; 6 - Empty exception table lo = $00

  numberOfParameters := byte[ methodBlockAdr + 0 ] >> jemFileCountAdjust
  numberOfLocals     := byte[ methodBlockAdr + 1 ] >> jemFileCountAdjust
  flags              := byte[ methodBlockAdr + 2 ]
  methodAdr          := byte[ methodBlockAdr + 3 ] << 8
  methodAdr          |= byte[ methodBlockAdr + 4 ]
  exceptionTable     := byte[ methodBlockAdr + 5 ] << 8
  exceptionTable     |= byte[ methodBlockAdr + 6 ]

  '               SP -->
  '       .----.        .----.          .----.          .----.
  '       | A  |        |    |          | SP |          | SP |
  ' SP -->| B  |        :    :          :    :          :    :
  '       `----'        |    |          | BP |          | BP |
  '                     :    :          :    :          :    :
  '                     |    |          | PC |          | PC |
  '                     |----|          |----|          |----|
  '                     | A  |    BP -->| A  |    BP -->| A  |
  '                     | B  |    SP -->| B  |          | B  |
  '                     `----'          `----'          |----|
  '                                                     | 0  |
  '                                                     | 0  |
  '                                               SP -->| 0  |
  '                                                     `----'
  '
  ' 1) Adjust SP to below parameters, shuffle parameters up
  ' 2) Insert stack frame, adjust BP and SP
  ' 3) Push zeroed local variables

  sp -= ( numberOfParameters << INTWORDS )
  
  if numberOfParameters <> 0

    frameStackSize := ( ( @sp - @pc ) >> 2 ) + 1 ' How many PushAddr's follow

    if not INT32
      if ChipVer <> 1           
        frameStackSize <<= 1
                                
    repeat i from 1 to numberOfParameters           
      Poke( sp, i + frameStackSize, Peek( sp, i ) )
                                                   
  repeat i from @sp to @pc step 4 ' Not -4 ! That's a Propeller Gotcha
    PushAddr( long[ i ] )
                  
  bp := sp + INTBYTES

  sp += ( numberOfParameters << INTWORDS )

  repeat numberOfLocals
    Push( 0 )

  et := exceptionTable + bytecodeStart
  
  case flags
  
    $00 :   pc := methodAdr + bytecodeStart
            mb := pc
            if hasClinit
              repeat while byte[ mb ] == jem#Jem_INVOKESTATIC
                mb += 3
              pc := mb ' ****** HACK **** This skips all "invokestatic <??.<clinit>>" at start of 'main'

    $80 :   if NativeMethod( methodAdr, numberOfParameters )
              ReturnFromMethodWithValue
            else
              ReturnFromMethod

    other : Crashed( String( "Unknown Method Table Flag value" ) )

PRI ReturnFromMethodWithValue

  '       .----.                  .----.
  '       | ?? |            BP -->| ?? |  <------.
  '       :    :                  :    :         |
  '       | ?? |                  | ?? |  <--.   |
  '       |----|                  |----|     |   |
  '       | SP |------.     SP -->| XX |     |   |
  '       :    :      |           `----'     |   |
  '       | BP |---.  |                      |   |
  '       :    :   |  `----------------------'   |
  '       | PC |   |                             |
  '       |----|   `-----------------------------'
  ' BP -->| A  |
  '       | B  |
  '       |----|
  ' SP -->| XX |
  '       `----'
  
  result := Pop
  ReturnFromMethod
  Push( result )
      
PRI ReturnFromMethod | i

  ' Note that it would not be essential to have pushed the SP in the current implementation as the SP
  ' is always the address preceding the stack frame and could therefore be determined as an offset from
  ' BP ( and also SP would point to where it should be when PC, MB and BP have been popped ) but this
  ' is designed to pre-empt stack frames having been allocated from the global heap at some time in the
  ' future.
  ' 
  '       .----.                  .----.
  '       | ?? |            BP -->| ?? |  <------.
  '       :    :                  :    :         |
  '       | ?? |            SP -->| ?? |  <--.   |
  '       |----|                  `----'     |   |
  '       | SP |-----------------------------'   |
  '       :    :                                 |
  '       | BP |---------------------------------'
  '       :    :
  '       | PC |
  '       |----|
  ' BP -->| A  |
  ' SP -->| B  |
  '       `----'

  sp := bp - INTBYTES

  repeat i from @pc to @sp step 4
    long[ i ] := PopAddr
    
' .-----------------------------------------------------------------------------------------------------.
' |     Call Jvm PASM Handler                                                                           |
' `-----------------------------------------------------------------------------------------------------'

PRI JvmPasm

  result := false
  
' .-----------------------------------------------------------------------------------------------------.
' |     Utility Code                                                                                    |
' `-----------------------------------------------------------------------------------------------------'

PRI ShowStack

  if stackAltered
    stackAltered := false
    tv.Out( $0D )
    tv.Hex( sp, 4 )
    tv.Out( " " )
    tv.Hex( ( ( sp - @stack ) >> INTWORDS ) + 1, 2)
    tv.out( " " )
    tv.out( " " )
    if  INT32
      ShowStackLong( 2 )
      ShowStackLong( 1 )
      ShowStackLong( 0 )
    else
      ShowStackWord( 5 )
      ShowStackWord( 4 )
      ShowStackWord( 3 )
      ShowStackWord( 2 )
      ShowStackWord( 1 )
      ShowStackWord( 0 )
  
PRI showStackLong( depth )

  tv.out( " " )
  if ( ( sp - @stack ) >> 2 ) => depth
    tv.Hex( long[ sp - (depth<<2) ], 8 )
  else
    tv.Str( String("--------") )

PRI showStackWord( depth )

  tv.out( " " )
  if ( ( sp - @stack ) >> 1 ) => depth
    tv.Hex( word[ sp - (depth<<1) ], 4 )
  else
    tv.Str( String("----") )

PRI ShowHeap | ptr, typ, bytLen, datPtr, nextPtr

  ptr := heap
  repeat while ptr < HEAP_BASE
    nextPtr := GetHeapInfo( ptr, @typ, @bytLen, 0 )
    ShowHeapEntry( ptr, typ, bytLen )
    ptr := nextPtr
    
PRI ShowHeapEntry( ptr, typ, bytLen )

  tv.Out( $0D )
  tv.Hex( ptr, 4 )
  tv.Out( "-" )
  if ( ChipVer == 1 ) or ( not BIGARRAYS )
    tv.Hex( ptr + bytLen + ( HEAP_HEAD_DATA << HEAP_WORD ) - 1, 4 )
  else
    tv.Hex( ptr + bytLen + ( HEAP_HEAD_DATA << HEAP_LONG ) - 1, 4 )
  tv.Out( " " )
  tv.Dec( typ )
  if typ < 10
    tv.Out( " " )
  tv.Out( " " )
  tv.Str( jem.TypeNamePtr(typ) )
  tv.Dec( BitSizePerType(typ) ) 
  tv.Out( "[" )
  tv.dec( bytLen >> BytesShiftPerType(typ) )
  tv.Out( "]" )  
  tv.Out( " " )
  if ( ChipVer == 1 ) or ( not BIGARRAYS )
    tv.Dec( HEAP_HEAD_DATA << HEAP_WORD )
  else
    tv.Dec( HEAP_HEAD_DATA << HEAP_LONG )
  tv.Out( "+" )
  tv.Dec( bytLen )
  tv.Str( String( " Bytes" ) )
  if typ & $80
    tv.Str( String( " DEAD" ) )
    
PRI BitSizePerType(typ)

  result := 8 << BytesShiftPerType(typ)

PRI CompactHeap | ptr, typ, bytLen, datPtr, nextPtr, nextTyp
{
  This should work but not tested and never called
  
  ptr := heap
  repeat while ptr < HEAP_BASE
    nextPtr := GetHeapInfo( ptr, @typ, @bytLen, @datPtr )
    if typ & jem#T_DELETED
      if ptr == heap
        heap := nextPtr
      else
        if nextPtr < HEAP_BASE
          GetHeapInfo( nextPtr, @nextTyp, 0, 0 )
          if nextTyp & jem#T_DELETED
            ' Combine the two
            nextPtr := GetHeapInfo( nextPtr, 0, 0, 0 )
            if ( ChipVer == 1 ) or ( not BIGARRAYS )
              word[ ptr + ( HEAP_HEAD_BLEN << HEAP_WORD ) ] := nextPtr - datPtr 
            else
              long[ ptr + ( HEAP_HEAD_BLEN << HEAP_LONG ) ] := nextPtr - datPtr  
            nextPtr := ptr
    ptr := nextPtr
}
               
PRI Crashed( txtPtr )

  tv.Out( $0D )
  tv.Str( txtPtr )

  repeat

PRI Pause_Ms( ms )

  waitcnt( CLKFREQ / 1000 * ms + CNT )
  
